from re        import sub, split
from .selector import Attribute, PathType, Selector

class MemoryBuffer(list):
    ''' Memory buffer for parser '''
    def flush(self):
        ''' Flush the buffer out as a string. '''
        flushed_string = []
        
        while self:
            flushed_string.append(self.pop())
        
        flushed_string.reverse()
        
        return ''.join(flushed_string)

class TokenizedSelectorBlock(object):
    ''' Template for Tokenized Selector Block '''
    def __init__(self):
        self.name           = None
        self.attributes     = set()
        self.pseudo_classes = set()

def selector(path):
    '''
    Parse a given *path* as the list of selector.
    '''
    
    selector = None
    path     = split(' ', sub('\s{2,}', ' ', path.strip()))
    
    previous_selector = None
    combinator        = None
    
    for p in path:
        if p in PathType.registered:
            combinator = PathType.registered.index(p)
            continue
        
        parsed_selector = selector_block(p)
        new_selector    = Selector(parsed_selector, combinator)
        combinator      = None
        
        if previous_selector:
            previous_selector.next(new_selector)
        
        previous_selector = new_selector
        
        if not selector:
            selector = new_selector
    
    return selector

def selector_block(raw_selector_block):
    '''
    Parse the *raw selector block*.
    
    .. note::
        This is a prototype code. The improvement is required.
    '''
    TOKENIZE_NAME         = 'name'
    TOKENIZE_ATTRIBUTE    = 'attr'
    TOKENIZE_PSEUDO_CLASS = 'pcls'
    
    token  = TokenizedSelectorBlock()
    memory = MemoryBuffer()
    current_step = TOKENIZE_NAME
    
    # Tokenized the raw block data.
    for ch in raw_selector_block:
        if ch == '[' and current_step == TOKENIZE_NAME:
            token.name   = memory.flush()
            current_step = TOKENIZE_ATTRIBUTE
        elif ch == ':' and current_step == TOKENIZE_NAME:
            token.name   = memory.flush()
            current_step = TOKENIZE_PSEUDO_CLASS
        elif ch == '[' and current_step == TOKENIZE_ATTRIBUTE:
            token.attributes.add(memory.flush())
        elif ch == ':' and current_step == TOKENIZE_ATTRIBUTE:
            token.attributes.add(memory.flush())
            current_step = TOKENIZE_PSEUDO_CLASS
        elif ch == ':' and current_step == TOKENIZE_PSEUDO_CLASS:
            token.pseudo_classes.add(memory.flush())
        
        memory.append(ch)
    
    # Finish the tokenization.
    if memory:
        leftover_buffer = memory.flush()
        
        if current_step == TOKENIZE_NAME:
            token.name = leftover_buffer
        elif current_step == TOKENIZE_ATTRIBUTE:
            token.attributes.add(leftover_buffer)
        elif current_step == TOKENIZE_PSEUDO_CLASS:
            token.pseudo_classes.add(leftover_buffer)
    
    return token



class Kami(list):
    """ The list of :class:`kotoba.kotoba.Kotoba`. """
    def __init__(self):
        self.__registered_nodes = []

    def children(self, selector=None):
        result = Kami()

        for item in self:
            result.extend(item.children(selector))

        return result

    def find(self, selector):
        result = Kami()

        for item in self:
            result.extend(item.find(selector))

        return result

    def data(self):
        output = []

        for kotoba in self:
            output.append(kotoba.data())

        try:
            return ''.join(output)
        except TypeError as e:
            if len(output) == 0:
                return None

            if len(output) == 1:
                return output[0]

            return output

    def append(self, kotoba):
        if kotoba.guid() in self.__registered_nodes:
            return

        self.__registered_nodes.append(kotoba.guid())

        super(Kami, self).append(kotoba)

    def extend(self, other_kami):
        for kotoba in other_kami:
            self.append(kotoba)


from codecs  import open as fopen
from os.path import exists

from .exception import *

def read(location):
    if not exists(location):
        raise FileNotFoundError
    
    with fopen(location) as fp:
        content = fp.read()
    
    fp.close()
    
    return content


import codecs
import json
import os
import re
from xml.dom.minidom   import parse, parseString

from .common    import is_string
from .driver    import JSONDriver, XMLDriver
from .kotoba    import Kotoba
from .exception import *

__all__ = ['Kotoba', 'load_from_file']

__version__ = (3, 2, 0)

def __load_xml(file_path):
    domDocument = parse(file_path)

    return XMLDriver(domDocument.documentElement)

def __load_json(file_path):
    with codecs.open(file_path) as f:
        obj = json.load(f)

    return JSONDriver(obj, 'root')

def load_from_file(file_path):
    """
    Load from the *filename*.

    :param file_path: the location of the data.

    :return: :class `kotoba.kotoba.Kotoba`: if the parser can parse the data.
    """
    if not os.path.exists(file_path):
        raise InvalidDataSourceError('The path {} is not found.'.format(file_path))

    if os.path.isdir(file_path):
        raise InvalidDataSourceError('The path {} is not a file.'.format(file_path))

    if re.search(r'\.json$', file_path, re.I):
        return Kotoba(__load_json(file_path))

    # default to XML
    return Kotoba(__load_xml(file_path))
